/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.nodediscovery;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbTrailTag;

import java.io.UnsupportedEncodingException;

import java.net.URLDecoder;

import javax.servlet.jsp.tagext.BodyTagSupport;


/**
 * <p>
 * This class provides a tag (only) implementation of node discovery that will allow the user to hardcode the url into
 * the body of this tag and specify the <b>title</b> (optional) in the attributes. These information is used to
 * construct the BreadCrumbNode.
 * </p>
 *
 * <p>
 * <b>Implementation notes:</b> This tag will be used by the application to specify the node url via the body of this
 * tag. This can be done by the application where the node url cannot be discovered by any implementations in this
 * component or if the application simply wishes to override or otherwise hardcode the url. This class cannot be used
 * automatically by the BreadCrumbTrailTag using the object factory because it requires user input, and therefore
 * doesn't implement the NodeDiscovery interface. Once the tag is processed, the current node is saved to the
 * BreadCrumbTrailTag.BREADCRUMB_CURRENT_NODE attribute using page scope.
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 */
public class InlineTextNodeDiscoveryTag extends BodyTagSupport {
    /**
     * <p>
     * Represents the title associated with this node discovery.  This variable is set initially to null, is mutable
     * and can be null but never an empty string.  A null title means that the title for the current node should
     * default to that found in the PathDiscovery data source. This variable is referenced by the associated
     * getter/setter methods.
     * </p>
     */
    private String title = null;

    /**
     * <p>
     * Constructs the InlineTextNodeDiscoveryTag. Simply calls the superclass constructor.
     * </p>
     */
    public InlineTextNodeDiscoveryTag() {
        super();
    }

    /**
     * <p>
     * Does nothing except returning EVAL_BODY_BUFFERED to tell the JSP engine to evaluate the body containing the URL.
     * </p>
     *
     * @return Always EVAL_BODY_BUFFERED to continue evaluating the body
     */
    public int doStartTag() {
        return EVAL_BODY_BUFFERED;
    }

    /**
     * <p>
     * Discovers the node url from the body content. The current node is constructed with the title attribute and the
     * url, and saved to the BreadCrumbTrailTag.BREADCRUMB_CURRENT_NODE attribute using page scope.
     * </p>
     *
     * @return Always SKIP_BODY to leave the processed body
     *
     * @throws BreadCrumbException if the body content is an empty (trim'd) string
     */
    public int doAfterBody() throws BreadCrumbException {
        // get the url from the body content
        String url = bodyContent.getString();

        if (url.trim().length() == 0) {
            throw new BreadCrumbException("Body content is empty");
        }

        try {
            // decode the url
            url = URLDecoder.decode(url, "utf-8");
        } catch (UnsupportedEncodingException e) {
            // should not happen
        }

        // save the current node in the attribute
        pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_CURRENT_NODE, new BreadCrumbNode(url, title));

        return SKIP_BODY;
    }

    /**
     * <p>
     * Returns the value of title attribute.
     * </p>
     *
     * @return The possibly null, non-empty string representing the title
     */
    public String getTitle() {
        return title;
    }

    /**
     * <p>
     * Sets the value of title attribute.
     * </p>
     *
     * @param title A possibly null, possibly empty (trim'd) string representing the title
     */
    public void setTitle(String title) {
        this.title = ((title != null) && (title.trim().length() == 0)) ? null : title;
    }

    /**
     * <p>
     * Resets all attribute member variables to their initial values.
     * </p>
     */
    public void release() {
        title = null;
        super.release();
    }
}
