/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.nodeformatter;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbTrailTag;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.jsp.tagext.TagSupport;


/**
 * <p>
 * This empty tag will format the node and style class using a template. A template is a separate file that will be
 * included in the output stream. This class will take the current node and style class, and define them as request
 * attributes when the template file is included, namely, BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE and
 * BreadCrumbTrailTag.BREADCRUMB_FORMAT_STYLE_CLASS. Those attributes can then be used by the template file to produce
 * the correct representation format.
 * </p>
 *
 * <p>
 * It supports the following attribute:
 *
 * <ul>
 * <li>
 * <b>template</b> (optional) - the template file that will be used when formatting this tag. If not specified, nothing
 * will be written to the output, which means the node is not rendered.
 * </li>
 * </ul>
 * </p>
 *
 * <p>
 * <b>Implementation notes:</b> This node formatter tag will be used in the body of a BreadCrumbTrailTag and will
 * format the current node and style class using a template file.
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 */
public class TemplateNodeFormatterTag extends TagSupport {
    /**
     * <p>
     * Represents the default template file that will be used when formatting this tag. This variable is set by the
     * constructor, is immutable and can be null or an empty (trim'd) string to specify no template is to be used (and
     * thus nothing is produced by this tag).
     * </p>
     */
    private final String defaultTemplate;

    /**
     * <p>
     * Represents the template file that will be used when formatting this tag. This variable is set by the
     * constructor, is mutable and can be null or an empty (trim'd) string to specify no template is to be used (and
     * thus nothing is produced by this tag). This variable is referenced by the associated getter/setter methods and
     * the doStartTag() method.
     * </p>
     */
    private String template;

    /**
     * <p>
     * Constructs the TemplateNodeFormatterTag with no template file.
     * </p>
     */
    public TemplateNodeFormatterTag() {
        this(null);
    }

    /**
     * <p>
     * Constructs the TemplateNodeFormatterTag with the given template file.
     * </p>
     *
     * @param template A possibly null, possibly empty (trim'd) string representing the template to use
     */
    public TemplateNodeFormatterTag(String template) {
        super();
        defaultTemplate = template;
        setTemplate(template);
    }

    /**
     * <p>
     * Returns the value of template attribute.
     * </p>
     *
     * @return A possibly null, possibly empty (trim'd) string representing the template to use
     */
    public String getTemplate() {
        return template;
    }

    /**
     * <p>
     * Sets the value of template attribute.
     * </p>
     *
     * @param template A possibly null, possibly empty (trim'd) string representing the template to use
     */
    public void setTemplate(String template) {
        this.template = template;
    }

    /**
     * <p>
     * This method will take the current node and style class, make them into request attributes and include the
     * template file.
     * </p>
     *
     * <p>
     * This method assumes that:
     *
     * <ul>
     * <li>
     * the BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE attribute holds the current node.
     * </li>
     * <li>
     * the BreadCrumbTrailTag.BREADCRUMB_FORMAT_STYLE_CLASS attribute holds the CSS style class for the current node if
     * it is available.
     * </li>
     * </ul>
     *
     * If template file is set and any of these is not met, a BreadCrumbException will be thrown.
     * </p>
     *
     * @return Always SKIP_BODY to skip the empty body
     *
     * @throws BreadCrumbException if some exception occurs during processing
     */
    public int doStartTag() throws BreadCrumbException {
        if ((template != null) && (template.trim().length() > 0)) {
            ServletRequest request = pageContext.getRequest();

            // set the format parameters to the request attributes
            request.setAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE, FormatterUtil.getFormatNode(pageContext));
            request.setAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_STYLE_CLASS,
                FormatterUtil.getFormatStyleClass(pageContext));

            // include the template file
            try {
                pageContext.include(template);
            } catch (ServletException e) {
                throw new BreadCrumbException("Cannot include the template file", e);
            } catch (IOException e) {
                throw new BreadCrumbException("Some I/O error occurs", e);
            }
        }
        return SKIP_BODY;
    }

    /**
     * <p>
     * Resets all attribute member variables to their initial values.
     * </p>
     */
    public void release() {
        template = defaultTemplate;
        super.release();
    }
}
