/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import com.topcoder.util.objectfactory.ObjectFactory;
import com.topcoder.util.objectfactory.impl.ConfigManagerSpecificationFactory;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.PathDiscovery;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.servlet.jsp.PageContext;


/**
 * <p>
 * This implementation of the PathDiscovery will attempt to create a static site map and then apply dynamic information
 * to that site map. The dynamic information can either be in matching nodes with different query strings or in new
 * nodes that will be assumed to map to the last known node.
 * </p>
 *
 * <p>
 * <b>Implementation notes:</b> This implementation will use a SiteDiscovery to discover the site map. That site map
 * will then be used to help determine where the current node belongs and what it's path back to the root node is.
 * This class will store the static site map to a session variable called CURRENT_SITEMAP, the reverse site map to
 * CURRENT_REVERSE_SITEMAP so that the xml source needs to be parsed only once, and the current path to CURRENT_PATH.
 * This class will only be created by the object factory when a path discovery object is needed by the
 * BreadCrumbTrailTag.
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe by being immutable.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 */
public class DynamicPathDiscovery implements PathDiscovery {
    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the current known path. This
     * constant will be used by this class to retrieve the current path. The current path will be of NodePair[] array
     * type representing the known path from the root (0th index) to the prior page (length-1 index).
     * </p>
     */
    public static final String CURRENT_PATH = DynamicPathDiscovery.class.getName() + ".currentPath";

    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the site map. This constant
     * will be used by this class to retrieve the site map. The site map will be a Map with NodeMatcher as the key and
     * a List of NodeMatchers as the value, where the list can be empty if there are no children of the parent.
     * </p>
     */
    public static final String CURRENT_SITEMAP = XmlPathDiscovery.class.getName() + ".siteMap";

    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the reverse site map. This
     * constant will be used by this class to retrieve the reverse site map. The site map will be a Map with
     * NodeMatcher as the key and a List of NodeMatchers as the value, where the list can be null if there are no
     * parents for the child.
     * </p>
     */
    public static final String CURRENT_REVERSE_SITEMAP = XmlPathDiscovery.class.getName() + ".reverseSiteMap";

    /**
     * <p>
     * The object factory used to create instance of SiteDiscovery class. This will initially be null, and lazily
     * initialized in the getPath(). Once initialized, the reference will not change.
     * </p>
     */
    private ObjectFactory objectFactory = null;

    /**
     * <p>
     * The site discovery used to discover the site map for finding the path from the current node to the root. This
     * will initially be null, and lazily initialized in the getPath(). Once initialized, the reference will not
     * change.
     * </p>
     */
    private SiteDiscovery siteDiscovery = null;

    /**
     * <p>
     * Constructs the DynamicPathDiscovery. Does nothing.
     * </p>
     */
    public DynamicPathDiscovery() {
    }

    /**
     * <p>
     * This method attempts to determine the path from the current node to the root and returns that path. It first
     * detects the situation when the user clicks on the trail (going backwards) or a link following the last page
     * well defined in the site map. If not, the path will be based on the shortest path from the current node to the
     * root based on the site map. This method will learn the current path from the current node dynamically.
     * </p>
     *
     * <p>
     * After this method is called, the original site map and reverse site map should be saved to the CURRENT_SITEMAP
     * and CURRENT_REVERSE_SITEMAP attributes respectively. The current path is also saved to the CURRENT_PATH
     * attribute.
     * </p>
     *
     * @param currentNode A non-null node representing the current page
     * @param pageContext A non-null page context to use
     *
     * @return An array of BreadCrumbNodes representing the path from root (0th index) to current node (nth-1 index),
     *         or null if path couldn't be found
     *
     * @throws IllegalArgumentException if any argument is null
     * @throws BreadCrumbException if an exception occurs discovering the page
     */
    public BreadCrumbNode[] getPath(BreadCrumbNode currentNode, PageContext pageContext)
        throws BreadCrumbException {
        if (currentNode == null) {
            throw new IllegalArgumentException("currentNode is null");
        }
        if (pageContext == null) {
            throw new IllegalArgumentException("pageContext is null");
        }

        try {
            // lazily initialize the objects for site discovery
            if (objectFactory == null) {
                objectFactory = new ObjectFactory(new ConfigManagerSpecificationFactory(getClass().getName()));
            }
            if (siteDiscovery == null) {
                siteDiscovery = (SiteDiscovery) objectFactory.createObject(SiteDiscovery.class);
            }
        } catch (Exception e) {
            throw new BreadCrumbException("Fails to initialize site discovery", e);
        }

        // get the site map and reverse site map from the attributes
        Map siteMap = ShortestPathFinder.getSiteMap(pageContext, CURRENT_SITEMAP, siteDiscovery);
        Map reverseSiteMap = ShortestPathFinder.getReverseSiteMap(pageContext, CURRENT_SITEMAP,
                CURRENT_REVERSE_SITEMAP, siteDiscovery);

        // get the current path from the attribute
        Object obj = pageContext.getAttribute(CURRENT_PATH, PageContext.SESSION_SCOPE);
        NodePair[] currentPath = null;

        if (obj instanceof NodePair[]) {
            currentPath = (NodePair[]) obj;
        }

        // check the current path backwards
        NodePair[] newPath = checkBackward(currentPath, currentNode);

        // check the current path forwards
        if (newPath == null) {
            newPath = checkForward(currentPath, currentNode, siteMap);
        }

        // compose the current path from the site map
        if (newPath == null) {
            newPath = checkShortestPath(currentPath, currentNode, reverseSiteMap);
        }

        // if the user visits the page from somewhere not in the site map, return null
        if (newPath == null) {
            return null;
        }

        // save the current path in the attribute
        pageContext.setAttribute(CURRENT_PATH, newPath, PageContext.SESSION_SCOPE);

        // convert the current path to an array of BreadCrumbNodes
        BreadCrumbNode[] nodes = new BreadCrumbNode[newPath.length];

        for (int i = 0; i < nodes.length; i++) {
            nodes[i] = newPath[i].getNode();
        }
        return nodes;
    }

    /**
     * <p>
     * Creates the new current path by copying the old elements with the specified length, and appending a new node
     * pair at the end.
     * </p>
     *
     * @param oldPath the old current path.
     * @param copyLength the number of old elements to copy.
     * @param lastNodePair the element to append at the end.
     *
     * @return the new current path.
     */
    private NodePair[] createCurrentPath(NodePair[] oldPath, int copyLength, NodePair lastNodePair) {
        NodePair[] newPath = new NodePair[copyLength + 1];

        // copy the old elements
        System.arraycopy(oldPath, 0, newPath, 0, copyLength);

        // append the last element
        newPath[copyLength] = lastNodePair;

        return newPath;
    }

    /**
     * <p>
     * Creates the node pair based on the information in the node matcher.
     * </p>
     *
     * @param nodeMatcher the node matcher to use.
     *
     * @return the created node pair.
     */
    private NodePair createNodePair(NodeMatcher nodeMatcher) {
        return new NodePair(new BreadCrumbNode(nodeMatcher.getUrl(), nodeMatcher.getTitle()), nodeMatcher);
    }

    /**
     * <p>
     * Checks the current path backwards with the current node. It traces the trails visited before. This will cover
     * the situations where the user clicked on the trail (which should match somewhere on the trail) or pressed the
     * back button (which should match the last node).
     * </p>
     *
     * <p>
     * The current path can be null if it is not available.
     * </p>
     *
     * @param currentPath the current path to use.
     * @param currentNode the node representing the current page.
     *
     * @return the new current path, or null if none can be found.
     */
    private NodePair[] checkBackward(NodePair[] currentPath, BreadCrumbNode currentNode) {
        if (currentPath == null) {
            return null;
        }

        // look for the matching node in the current path backwards
        for (int i = currentPath.length - 1; i >= 0; i--) {
            BreadCrumbNode node = currentPath[i].getNode();
            NodeMatcher nodeMatcher = currentPath[i].getNodeMatcher();

            if (((nodeMatcher == null) && node.getUrl().equals(currentNode.getUrl()))
                    || ((nodeMatcher != null) && nodeMatcher.matches(currentNode))) {
                // simply add the old elements up to the matching node,
                // and update the last element
                return createCurrentPath(currentPath, i, new NodePair(currentNode, nodeMatcher));
            }
        }

        // all node matchers have been exhausted
        return null;
    }

    /**
     * <p>
     * Checks the current path forwards with the current node. It traces the trails from the last visited page. This
     * covers the situation where the user went to a place that is well defined (in the site map).
     * </p>
     *
     * <p>
     * The current path can be null if it is not available.
     * </p>
     *
     * @param currentPath the current path to use.
     * @param currentNode the node representing the current page.
     * @param siteMap A non-null map of parent NodeMatcher to a List of NodeMatcher children
     *
     * @return the new current path, or null if none can be found.
     */
    private NodePair[] checkForward(NodePair[] currentPath, BreadCrumbNode currentNode, Map siteMap) {
        if (currentPath == null) {
            return null;
        }

        // get the last node matcher from the current path
        NodeMatcher nodeMatcher = currentPath[currentPath.length - 1].getNodeMatcher();

        if (nodeMatcher == null) {
            return null;
        }

        // get the children of the last node matcher
        List children = (List) siteMap.get(nodeMatcher);

        if (children == null) {
            return null;
        }

        // look for the matching node in the current path forwards
        for (Iterator i = children.iterator(); i.hasNext();) {
            nodeMatcher = (NodeMatcher) i.next();

            if (nodeMatcher.matches(currentNode)) {
                // simply append the current node to the current path
                return createCurrentPath(currentPath, currentPath.length, new NodePair(currentNode, nodeMatcher));
            }
        }

        // all node matchers have been exhausted
        return null;
    }

    /**
     * <p>
     * Checks the shortest path with the current node. To choose the best shortest path, it combines information
     * obtained from the current path if available.
     * </p>
     *
     * <p>
     * The current path can be null if it is not available. This method should be called if it fails in checking the
     * current path backwards and forwards.
     * </p>
     *
     * @param currentPath the current path to use.
     * @param currentNode the node representing the current page.
     * @param reverseSiteMap A non-null map of child NodeMatcher to a List of NodeMatcher parents
     *
     * @return the new current path, or null if none can be found.
     */
    private NodePair[] checkShortestPath(NodePair[] currentPath, BreadCrumbNode currentNode, Map reverseSiteMap) {
        // find the shortest paths using the site map
        ShortestPathFinder finder = new ShortestPathFinder(reverseSiteMap);
        NodeMatcher[][] paths = finder.findShortestPaths(currentNode);

        // if no shortest path is found
        if (paths == null) {
            if (currentPath == null) {
                // if there is no current path, return null indicating an unknown path
                return null;
            } else {
                // otherwise, simply append the current node to the current path
                return createCurrentPath(currentPath, currentPath.length, new NodePair(currentNode, null));
            }
        }

        // choose the first shortest path by default
        NodeMatcher[] nodeMatchers = paths[0];
        int matchedCount = 0;

        // if current path is available, pick the best shortest path that has the most commonality
        if (currentPath != null) {
            for (int i = 0; i < paths.length; i++) {
                int count = 0;

                // compare the node matchers of the two paths
                for (; (count < paths[i].length) && (count < currentPath.length); count++) {
                    if (!paths[i][count].equals(currentPath[count].getNodeMatcher())) {
                        break;
                    }
                }
                if (count > matchedCount) {
                    nodeMatchers = paths[i];
                    matchedCount = count;
                }
            }
        }

        NodePair[] newPath = new NodePair[nodeMatchers.length];

        // copy the matched elements from the current path
        if (matchedCount > 0) {
            System.arraycopy(currentPath, 0, newPath, 0, matchedCount);
        }

        // fill the rest with the elements from the best shortest path
        for (int i = matchedCount; i < newPath.length; i++) {
            newPath[i] = createNodePair(nodeMatchers[i]);
        }
        return newPath;
    }
}
