/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.PathDiscovery;
import com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.nodematchers.RegexNodeMatcher;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.xml.sax.SAXException;

import java.io.ByteArrayInputStream;
import java.io.IOException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.jsp.PageContext;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;


/**
 * <p>
 * This implementation of the PathDiscovery and SiteDiscovery will attempt to create a static site map from either a
 * filename or string source, both of which is an XML document. This class will parse the document to create a static
 * map of RegexNodeMatchers. If used as a PathDiscovery, this provides a static view of the site.
 * </p>
 *
 * <p>
 * This is an example of the XML document:
 * <pre>
 * &lt;bc_node title="MyMain" url="/" pattern="/"&gt;
 *   &lt;bc_node title="Forums" url="/forums" pattern="/forums"&gt;
 *     &lt;bc_node title="Help" url="/forums" pattern="/forums?id.*"&gt;
 *       &lt;bc_node title="About" url="/forums" pattern="/forums?id=about"/&gt;
 *     &lt;/bc_node&gt;
 *   &lt;/bc_node&gt;
 *   &lt;bc_node title="Topics" url="/topics" pattern="/topics"/&gt;
 * &lt;/bc_node&gt;
 *  &lt;bc_node title="AltMain" url="/main" pattern="/"&gt;
 *   &lt;bc_node title="Help" url="/forums" pattern="/forums?id.*"/&gt;
 * &lt;/bc_node&gt;
 * </pre>
 * All elements should have the name "bc_node" to denote the bread crumb node. Note that in the above, the "Help" node
 * appears two times. However only one node (any of them) is specified with child nodes. It is assumed that the other
 * occurrences (without child nodes) follow the same structure.
 * </p>
 *
 * <p>
 * <b>Implementation notes:</b> This implementation will be used as either a non-dynamic path discovery operation to be
 * used by the BreadCrumbTrailTag or can be used as a SiteDiscovery for any other PathDiscovery implementation. The
 * BreadCrumbTrailTag will either construct this class directly (if the pathOverride attribute is specified) or
 * automatically via the object factory. This class will store the static site map to a session variable called
 * CURRENT_SITEMAP, and reverse site map to CURRENT_REVERSE_SITEMAP so that the xml source needs to be parsed only
 * once.
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe by being immutable.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 */
public class XmlPathDiscovery implements SiteDiscovery, PathDiscovery {
    /**
     * <p>
     * The name of the element in the xml file corresponding to a bread crumb node.
     * </p>
     */
    private static final String ELEMENT_NAME = "bc_node";

    /**
     * <p>
     * The name of the title attribute corresponding to the node matcher's title.
     * </p>
     */
    private static final String TITLE_ATTRIBUTE = "title";

    /**
     * <p>
     * The name of the url attribute corresponding to the node matcher's url.
     * </p>
     */
    private static final String URL_ATTRIBUTE = "url";

    /**
     * <p>
     * The name of the pattern attribute corresponding to the node matcher's regex pattern.
     * </p>
     */
    private static final String PATTERN_ATTRIBUTE = "pattern";

    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the site map. This constant
     * will be used by this class to retrieve the site map. The site map will be a Map with NodeMatcher as the key and
     * a List of NodeMatchers as the value, where the list can be empty if there are no children of the parent.
     * </p>
     */
    public static final String CURRENT_SITEMAP = XmlPathDiscovery.class.getName() + ".siteMap";

    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the reverse site map. This
     * constant will be used by this class to retrieve the reverse site map. The site map will be a Map with
     * NodeMatcher as the key and a List of NodeMatchers as the value, where the list can be null if there are no
     * parents for the child.
     * </p>
     */
    public static final String CURRENT_REVERSE_SITEMAP = XmlPathDiscovery.class.getName() + ".reverseSiteMap";

    /**
     * <p>
     * Represents the source of the site data. The source will be set by the constructor, is immutable and will never
     * be null or an empty (trim'd) string. This variable is referenced by the getSiteMap() method.
     * </p>
     */
    private final String source;

    /**
     * <p>
     * Represents whether the source represents a file (true) or a raw xml string (false). The isFile will be set by
     * the constructor and is immutable. This variable is referenced by the getSiteMap() method.
     * </p>
     */
    private final boolean isFile;

    /**
     * <p>
     * Constructs the XmlPathDiscovery from the given filename. This is equivalent to calling
     * XmlPathDiscovery(filename, true).
     * </p>
     *
     * @param filename A non-null, non-empty (trim'd) string representing the file to use
     *
     * @throws IllegalArgumentException if the filename is null or an empty (trim'd) string
     */
    public XmlPathDiscovery(String filename) {
        this(filename, true);
    }

    /**
     * <p>
     * Constructs the XmlPathDiscovery from the given source and whether that source is a direct XML string or a
     * filename.
     * </p>
     *
     * @param source A non-null, non-empty (trim'd) string representing the source
     * @param isFile Whether the source is a file (true) or a raw xml string (false)
     *
     * @throws IllegalArgumentException if the source is null or an emtpy (trim'd) string
     */
    public XmlPathDiscovery(String source, boolean isFile) {
        if (source == null) {
            throw new IllegalArgumentException("source is null");
        }
        if (source.trim().length() == 0) {
            throw new IllegalArgumentException("source is empty string");
        }
        this.source = source;
        this.isFile = isFile;
    }

    /**
     * <p>
     * Returns a map of NodeMatcher keys with a List of NodeMatcher elements as the value. The key NodeMatcher
     * represents the parent and the value List represents the child values. The value can be null or an empty list to
     * represent no child elements.
     * </p>
     *
     * @return A non-null map of parent NodeMatcher to a List of NodeMatcher children
     *
     * @throws BreadCrumbException if parsing of data source fails
     */
    public Map getSiteMap() throws BreadCrumbException {
        Document document = null;

        try {
            // use DOM for parsing
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();

            if (isFile) {
                // parse the XML file from its filename
                document = builder.parse(source);
            } else {
                // or parse the XML string directly
                document = builder.parse(new ByteArrayInputStream(source.getBytes()));
            }
        } catch (ParserConfigurationException e) {
            throw new BreadCrumbException("Fails to create the parser", e);
        } catch (SAXException e) {
            throw new BreadCrumbException("Fails to parse the data source", e);
        } catch (IOException e) {
            throw new BreadCrumbException("Some I/O error occurs in parsing", e);
        }

        // build the mapping from the children of root element
        Element root = (Element) document.getDocumentElement();
        Map siteMap = new HashMap();

        for (Node childNode = root.getFirstChild(); childNode != null; childNode = childNode.getNextSibling()) {
            if ((childNode.getNodeType() != Node.ELEMENT_NODE) || !childNode.getNodeName().equals(ELEMENT_NAME)) {
                continue;
            }
            buildSiteMap(childNode, createNodeMatcher((Element) childNode), siteMap);
        }
        return siteMap;
    }

    /**
     * <p>
     * This method attempts to determine the shortest path from the current node to the root and returns that path
     * based on the current site map. If there are more than one shortest paths, an arbitrary one will be chosen. The
     * last node will be set to the current node, except that the title will be based on the site map if it is not
     * available from the current node.
     * </p>
     *
     * <p>
     * After this method is called, the original site map and reverse site map should be saved to the CURRENT_SITEMAP
     * and CURRENT_REVERSE_SITEMAP attributes respectively.
     * </p>
     *
     * @param currentNode A non-null node representing the current page
     * @param pageContext A non-null page context to use
     *
     * @return An array of BreadCrumbNodes representing the path from root (0th index) to current node (nth-1 index),
     *         or null if path couldn't be found
     *
     * @throws IllegalArgumentException if any argument is null
     * @throws BreadCrumbException if an exception occurs discovering the page
     */
    public BreadCrumbNode[] getPath(BreadCrumbNode currentNode, PageContext pageContext)
        throws BreadCrumbException {
        if (currentNode == null) {
            throw new IllegalArgumentException("currentNode is null");
        }
        if (pageContext == null) {
            throw new IllegalArgumentException("pageContext is null");
        }

        // get the reverse site map from the attributes
        Map reverseSiteMap = ShortestPathFinder.getReverseSiteMap(pageContext, CURRENT_SITEMAP,
                CURRENT_REVERSE_SITEMAP, this);

        // find the shortest paths from the reverse site map
        ShortestPathFinder finder = new ShortestPathFinder(reverseSiteMap);
        NodeMatcher[][] paths = finder.findShortestPaths(currentNode);

        // if no shortest path is found, return null
        if (paths == null) {
            return null;
        }

        // pick the first shortest path if there are many
        NodeMatcher[] nodeMatchers = paths[0];
        BreadCrumbNode[] nodes = new BreadCrumbNode[nodeMatchers.length];
        int last = nodes.length - 1;

        // compose the path from the node matchers
        for (int i = 0; i < last; i++) {
            nodes[i] = new BreadCrumbNode(nodeMatchers[i].getUrl(), nodeMatchers[i].getTitle());
        }

        // the last node in the path should be based on the current node (except the title if not available)
        if (currentNode.getTitle() == null) {
            nodes[last] = new BreadCrumbNode(currentNode.getUrl(), nodeMatchers[last].getTitle());
        } else {
            nodes[last] = currentNode;
        }

        return nodes;
    }

    /**
     * <p>
     * Builds the site map starting from the given parentNode. The parent corresponds to the NodeMatcher produced from
     * the parentNode. The results are saved into the given siteMap.
     * </p>
     *
     * @param parentNode the parent node to build the site map from.
     * @param parent the node matcher produced by the parent node.
     * @param siteMap the site map where the results are stored into.
     *
     * @throws BreadCrumbException if any error occurs building the site map.
     */
    private void buildSiteMap(Node parentNode, NodeMatcher parent, Map siteMap)
        throws BreadCrumbException {
        List children = (List) siteMap.get(parent);

        if (children == null) {
            // here the parent appears for the first time,
            // map the parent (NodeMatcher) to the children (List of NodeMatchers)
            children = new ArrayList();
            siteMap.put(parent, children);
        } else {
            // here the parent appears again
            if (parentNode.hasChildNodes() && !children.isEmpty()) {
                throw new BreadCrumbException("Duplicate parent nodes should not have child nodes");
            }
        }

        for (Node childNode = parentNode.getFirstChild(); childNode != null; childNode = childNode.getNextSibling()) {
            if ((childNode.getNodeType() != Node.ELEMENT_NODE) || !childNode.getNodeName().equals(ELEMENT_NAME)) {
                continue;
            }

            // create the RegexNodeMatcher for each child from the attributes
            NodeMatcher child = createNodeMatcher((Element) childNode);

            // add the node matcher to the children list
            children.add(child);

            // recursively build the relationship for each child
            buildSiteMap(childNode, child, siteMap);
        }
    }

    /**
     * <p>
     * Creates the node matcher based on the required attributes of the given element.
     * </p>
     *
     * @param element the element containing the attributes.
     *
     * @return the created node matcher.
     *
     * @throws BreadCrumbException if any required attribute is missing, or regex pattern is invalid.
     */
    private NodeMatcher createNodeMatcher(Element element)
        throws BreadCrumbException {
        try {
            return new RegexNodeMatcher(element.getAttribute(TITLE_ATTRIBUTE), element.getAttribute(URL_ATTRIBUTE),
                element.getAttribute(PATTERN_ATTRIBUTE));
        } catch (IllegalArgumentException e) {
            throw new BreadCrumbException("Missing required attribute", e);
        }
    }
}
