/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.nodematchers;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.AbstractNodeMatcher;

import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;


/**
 * <p>
 * This implementation of a NodeMatcher will provide node matching services given a specific regex pattern. This class
 * will return true when the passed node's url matches the specified pattern.
 * </p>
 *
 * <p>
 * <b>Implementation notes:</b> This class will be created by an XmlPathDiscovery object when it's creating the site
 * map. This class will then be used by both the XmlPathDiscovery and DynamicPathDiscovery objects to determine the
 * path(s) to the root node(s).
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe by being immutable.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 */
public class RegexNodeMatcher extends AbstractNodeMatcher {
    /**
     * <p>
     * Represents the regex pattern that will be used to determine if the node matches.  This variable is set  by the
     * constructor, is immutable and can never be null or an empty (trim'd) string.  The pattern will be returned in
     * the getNodePattern() getter method.
     * </p>
     */
    private final String nodePattern;

    /**
     * <p>
     * Represents the compiled version of the regex pattern.  This variable is set  by the constructor, is immutable
     * and can never be null.  The pattern will be used in the matches() method.
     * </p>
     */
    private final Pattern pattern;

    /**
     * <p>
     * Constructs the RegexNodeMatcher from the given title, url and pattern.
     * </p>
     *
     * @param title A non-null, non-empty (trim'd) string describing the default title of the node
     * @param url A non-null, non-empty (trim'd) string representing the default url of the node
     * @param nodePattern A non-null, non-empty (trim'd) string representing the regex pattern to apply
     *
     * @throws IllegalArgumentException if any argument is null or an empty (trim'd) string
     * @throws BreadCrumbException if the nodePattern is an invalid regex.
     */
    public RegexNodeMatcher(String title, String url, String nodePattern)
        throws BreadCrumbException {
        super(title, url);

        if (nodePattern == null) {
            throw new IllegalArgumentException("nodePattern is null");
        }
        if (nodePattern.trim().length() == 0) {
            throw new IllegalArgumentException("nodePattern is empty string");
        }
        try {
            pattern = Pattern.compile(nodePattern);
        } catch (PatternSyntaxException e) {
            throw new BreadCrumbException("nodePattern is an invalid regex", e);
        }
        this.nodePattern = nodePattern;
    }

    /**
     * <p>
     * Returns true if the node's url matches the specified pattern.
     * </p>
     *
     * @param node A possibly null node to test
     *
     * @return True if the node matches the regex pattern, false otherwise
     */
    public boolean matches(BreadCrumbNode node) {
        return (node == null) ? false : pattern.matcher(node.getUrl()).matches();
    }

    /**
     * <p>
     * Simply the getter method for the node pattern string.
     * </p>
     *
     * @return A non-null, non-empty (trim'd) string representing the regex pattern used to match against nodes
     */
    public String getNodePattern() {
        return nodePattern;
    }

    /**
     * <p>
     * Returns true if the given object matches this node matcher. Equality holds only if the object is of type
     * RegexNodeMatcher and the nodePattern, title and url all match.
     * </p>
     *
     * @param o A possibly null object to test for equality
     *
     * @return True if the object is of type RegexNodeMatcher with a matching nodePattern, title and url
     */
    public boolean equals(Object o) {
        if (!(o instanceof RegexNodeMatcher)) {
            return false;
        }

        RegexNodeMatcher matcher = (RegexNodeMatcher) o;

        return nodePattern.equals(matcher.nodePattern) && getTitle().equals(matcher.getTitle())
            && getUrl().equals(matcher.getUrl());
    }

    /**
     * <p>
     * Returns the hash code for this object. This takes the node pattern, title and url into account.
     * </p>
     *
     * @return The hashcode of the object
     */
    public int hashCode() {
        return (nodePattern.hashCode() + 29) ^ (getTitle().hashCode() + 29) ^ getUrl().hashCode();
    }
}
