/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.nodeformatter;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbTrailTag;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.apache.cactus.JspTestCase;
import org.apache.cactus.WebResponse;


/**
 * Unit tests for TemplateNodeFormatterTag implementation.
 *
 * @author TCSDEVELOPER
 * @version 2.0
 */
public class TemplateNodeFormatterTagTest extends JspTestCase {
    /**
     * The template of the TemplateNodeFormatterTag.
     */
    private static final String TEMPLATE = "/template.txt";

    /**
     * The TemplateNodeFormatterTag instance to test against.
     */
    private TemplateNodeFormatterTag tag = null;

    /**
     * The node set to the BREADCRUMB_FORMAT_NODE attribute.
     */
    private BreadCrumbNode node = null;

    /**
     * The style class set to the BREADCRUMB_FORMAT_STYLE_CLASS attribute.
     */
    private String styleClass = null;

    /**
     * Creates a test suite for the tests in this test case.
     *
     * @return a TestSuite for this test case
     */
    public static Test suite() {
        TestSuite suite = new TestSuite(TemplateNodeFormatterTagTest.class);

        return suite;
    }

    /**
     * Prepares a TemplateNodeFormatterTag instance for testing. Also sets the BREADCRUMB_FORMAT_NODE and
     * BREADCRUMB_FORMAT_STYLE_CLASS attributes in page scope.
     */
    protected void setUp() {
        tag = new TemplateNodeFormatterTag();
        tag.setPageContext(pageContext);
        tag.setTemplate(TEMPLATE);

        node = new BreadCrumbNode("url", "title");
        pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE, node);

        styleClass = "styleClass";
        pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_STYLE_CLASS, styleClass);
    }

    /**
     * Test of constructor with no argument. Verifies if the template is set to null.
     */
    public void testConstructor_NoArg() {
        tag = new TemplateNodeFormatterTag();
        assertNull("Fails to set template", tag.getTemplate());
    }

    /**
     * Test of constructor with one argument. Verifies if the template is set as specified.
     */
    public void testConstructor_OneArg() {
        tag = new TemplateNodeFormatterTag(TEMPLATE);
        assertEquals("Fails to set template", TEMPLATE, tag.getTemplate());
    }

    /**
     * Test of getTemplate and setTemplate methods. Verifies it by getting the template back after setting it.
     */
    public void testGetSetTemplate() {
        tag.setTemplate(TEMPLATE);
        assertEquals("Returns an incorrect template", TEMPLATE, tag.getTemplate());
    }

    /**
     * Test of doStartTag method with no template specified. Should return SKIP_BODY.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testDoStartTag_WithoutTemplate() throws Exception {
        // make sure the template is not set
        tag.setTemplate(null);

        assertEquals("SKIP_BODY is not returned", tag.SKIP_BODY, tag.doStartTag());
    }

    /**
     * Verifies if nothing is written to the output.
     *
     * @param theResponse the response object to check the output string.
     */
    public void endDoStartTag_WithoutTemplate(WebResponse theResponse) {
        assertEquals("Prints an incorrect output", 0, theResponse.getText().length());
    }

    /**
     * Test of doStartTag method with no BREADCRUMB_FORMAT_NODE attribute. Expects BreadCrumbException.
     */
    public void testDoStartTag_NoNodeAttr() {
        // make sure the attribute is not set
        pageContext.removeAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE);

        try {
            tag.doStartTag();
            fail("Does start tag processing with no BREADCRUMB_FORMAT_NODE attribute");
        } catch (BreadCrumbException e) {
            // good
        }
    }

    /**
     * Test of doStartTag method with bad BREADCRUMB_FORMAT_NODE attribute (non-BreadCrumbNode type). Expects
     * BreadCrumbException.
     */
    public void testDoStartTag_BadNodeAttr() {
        // make sure the attribute is bad
        pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE, "string");

        try {
            tag.doStartTag();
            fail("Does start tag processing with bad BREADCRUMB_FORMAT_NODE attribute");
        } catch (BreadCrumbException e) {
            // good
        }
    }

    /**
     * Test of doStartTag method with bad BREADCRUMB_FORMAT_STYLE_CLASS attribute (non-String type). Expects
     * BreadCrumbException.
     */
    public void testDoStartTag_BadStyleClassAttr() {
        // make sure the attribute is bad
        pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_STYLE_CLASS, new Integer(0));

        try {
            tag.doStartTag();
            fail("Does start tag processing with bad BREADCRUMB_FORMAT_STYLE_CLASS attribute");
        } catch (BreadCrumbException e) {
            // good
        }
    }

    /**
     * Test of doStartTag method with template specified. Should return SKIP_BODY. Verifies if the request has the
     * BREADCRUMB_FORMAT_NODE attribute set to the node, and the BREADCRUMB_FORMAT_STYLE_CLASS attribute set to the
     * style class.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testDoStartTag_WithTemplate() throws Exception {
        assertEquals("SKIP_BODY is not returned", tag.SKIP_BODY, tag.doStartTag());
        assertEquals("Fails to set the node attribute", node,
            request.getAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE));
        assertEquals("Fails to set the style class attribute", styleClass,
            request.getAttribute(BreadCrumbTrailTag.BREADCRUMB_FORMAT_STYLE_CLASS));
    }

    /**
     * Verifies if the template is written to the output. This proves that the template file was included in the page
     * context.
     *
     * @param theResponse the response object to check the output string.
     */
    public void endDoStartTag_WithTemplate(WebResponse theResponse) {
        assertEquals("Prints an incorrect output", "template", theResponse.getText());
    }

    /**
     * Test of release method. Verifies if data members are reset to their initial values.
     */
    public void testRelease() {
        tag.release();
        assertNull("Fails to reset template", tag.getTemplate());
    }
}
